-------------------------------------------------------------------------------
-- Title      : MOUSEQ (Uncomplicated MOnitoring SEQuencer)
-- Project    : Distributed I/O Tier
-------------------------------------------------------------------------------
-- File       : mouseq.vhd
-- Author     : Christos Gentsos <christos.gentsos@cern.ch>
-- Company    : CERN (BE-CO-HT)
-- Standard   : VHDL
-------------------------------------------------------------------------------
-- Description:
--
-- This is the main file for the mouseq IP core. It instantiates an program
-- memory, a simple processor, and an i2c controller and integrates them such
-- that monitoring tasks are made much simpler.
-------------------------------------------------------------------------------
--
-- Copyright (c) 2019 CERN
--
-- Copyright and related rights are licensed under the Solderpad Hardware
-- License, Version 0.51 (the "License") (which enables you, at your option,
-- to treat this file as licensed under the Apache License 2.0); you may not
-- use this file except in compliance with the License. You may obtain a copy
-- of the License at http://solderpad.org/licenses/SHL-0.51.
-- Unless required by applicable law or agreed to in writing, software,
-- hardware and materials distributed under this License is distributed on an
-- "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
-- or implied. See the License for the specific language governing permissions
-- and limitations under the License.
--
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use work.mouseq_pkg.all;

entity mouseq is
  generic (
    CLK_PERIOD : natural := 10_000);      -- clock period in ps
  port (
    clk_i      : in std_logic;
    rst_n_i    : in std_logic;

    -- external trigger
    ext_trig_i : in std_logic;

    -- I/O regs
    regs_in_i  : in std_logic_vector(IWORDS*WORDWIDTH-1 downto 0);
    regs_out_o : out std_logic_vector(OWORDS*WORDWIDTH-1 downto 0);

    -- i2c bus
    scl_o      : out std_logic;
    scl_oen    : out std_logic;
    scl_i      : in std_logic;
    sda_o      : out std_logic;
    sda_oen    : out std_logic;
    sda_i      : in std_logic
    );
end mouseq;

architecture behavioral of mouseq is

  -- initially put to better measure resources but it seems to be giving
  -- better results like that, leaving it on.
  attribute syn_hier : string;
  attribute syn_hier of behavioral : architecture is "hard";

  attribute syn_radhardlevel : string;
  attribute syn_radhardlevel of behavioral: architecture is "none";

  signal pmem_fetch : std_logic;
  signal pmem_dv : std_logic;
  signal pmem_dout : std_logic_vector(MEMWIDTH-1 downto 0);

  -- I/O might be registered or no, declare signals to use
  signal regs_in  : std_logic_vector(IWORDS*WORDWIDTH-1 downto 0);
  signal regs_out : std_logic_vector(OWORDS*WORDWIDTH-1 downto 0);

  -- signals for i2c master component
  signal rst : std_logic;
  signal i2c_en      : std_logic;
  signal i2c_start   : std_logic;
  signal i2c_stop    : std_logic;
  signal i2c_rd      : std_logic;
  signal i2c_wr      : std_logic;
  signal i2c_txack   : std_logic;
  signal i2c_busy    : std_logic;
  signal i2c_lost    : std_logic;
  signal i2c_txbyte  : std_logic_vector(7 downto 0);
  signal i2c_rxbyte  : std_logic_vector(7 downto 0);
  signal i2c_done    : std_logic;
  signal i2c_rxack_n : std_logic;

  -- processor opcodes
  constant c_op_WRITE1    : std_logic_vector(3 downto 0) := "0010";
  constant c_op_WRITE2    : std_logic_vector(3 downto 0) := "0011";
  constant c_op_READ1     : std_logic_vector(3 downto 0) := "0100";
  constant c_op_READ2     : std_logic_vector(3 downto 0) := "0101";
  constant c_op_STOP      : std_logic_vector(3 downto 0) := "0110";
  constant c_op_GOTO      : std_logic_vector(3 downto 0) := "1001";
  constant c_op_WAIT      : std_logic_vector(3 downto 0) := "1010";
  constant c_op_WAIT_TRIG : std_logic_vector(3 downto 0) := "1011";
  constant c_op_MOV       : std_logic_vector(3 downto 0) := "1100";
  constant c_op_SYNCIO    : std_logic_vector(3 downto 0) := "1101";
  constant c_op_LOAD      : std_logic_vector(3 downto 0) := "1110";
  constant c_op_INC       : std_logic_vector(3 downto 0) := "1111";
  constant c_op_NOP       : std_logic_vector(3 downto 0) := "0000";

  -- program counter
  signal pc : unsigned(MEMABITS-1 downto 0);
  signal pc_slv : std_logic_vector(MEMABITS-1 downto 0);

  -- FSM states (description in the code)
  type state_t is (
    IDLE,
    STARTUP,
    FETCH,
    FETCH_av,
    FETCH_dv,
    DECODE,
    FETCH_WRITE_av,
    FETCH_WRITE_dv,
    FETCH_READ_av,
    FETCH_READ_dv,
    EXEC_WRITE,
    EXEC_WRITE_REG,
    EXEC_WRITE_DATA,
    EXEC_WRITE_CHECK_SIZE,
    EXEC_WRITE_WAIT_ADJ_UPD,
    EXEC_WRITE_FETCH_NEXT_DATA,
    EXEC_WRITE_WAIT_NEXT_DATA,
    EXEC_WRITE_SEND_STOP,
    EXEC_READ,
    EXEC_READ_PNT,
    EXEC_READ_PNT_WAIT,
    EXEC_READ_PNT_GO,
    EXEC_READ_REG,
    EXEC_READ_NOTFIRST,
    EXEC_READ_DATA,
    EXEC_READ_STORE_DATA,
    EXEC_READ_CHECK_SIZE,
    EXEC_READ_WAIT_ACK_DONE,
    EXEC_READ_WAIT_ADJ_UPD,
    EXEC_READ_SEND_STOP,
    EXEC_STOP,
    EXEC_STOP_2,
    EXEC_GOTO,
    EXEC_WAIT,
    EXEC_WAIT_2,
    EXEC_WAIT_TRIG,
    EXEC_MOV,
    EXEC_MOV_WAIT_R,
    EXEC_MOV_WAIT_W,
    FETCH_LOAD_av,
    FETCH_LOAD_dv,
    EXEC_LOAD,
    EXEC_LOAD_WAIT,
    EXEC_INC,
    EXEC_INC_WAIT_R,
    EXEC_INC_INC,
    EXEC_INC_WAIT_W,
    EXEC_SYNCIO,
    EXEC_SYNCIO_READIN_WAIT,
    EXEC_SYNCIO_WRITEOUT,
    EXEC_SYNCIO_WRITEOUT_WAIT
    );

  signal state : state_t;

  -- (optional) watchdog signals
  type watchdog_state_t is (
    WD_STATE_CHANGE,
    WD_STATE_SAME,
    WD_RESET
    );
  signal watchdog_state : watchdog_state_t;
  signal watchdog_cnt : unsigned(WATCHDOG_BITS-1 downto 0);
  signal state_r : state_t;

  -- watchdog reset start address self-scrubbing register
  signal start_at_reset_addr1 : std_logic;
  signal start_at_reset_addr2 : std_logic;
  signal start_at_reset_addr3 : std_logic;
  signal start_at_reset_addr_vote_out : std_logic;

  -- preserve the three copies
  attribute syn_preserve : boolean;
  attribute syn_preserve of start_at_reset_addr1 : signal is true;
  attribute syn_preserve of start_at_reset_addr2 : signal is true;
  attribute syn_preserve of start_at_reset_addr3 : signal is true;

  -- 18-bit command word
  signal cmd_word : std_logic_vector(MEMWIDTH-1 downto 0);

  -- second command word, for write and read operations
  signal cmd_word_2 : std_logic_vector(MEMWIDTH-1 downto 0);

  -- command opcode
  signal cmd_opcode : std_logic_vector(3 downto 0);

  -- various aliases for different command arguments
  signal cmd_local_reg_addr_or_literal : std_logic_vector(7 downto 0);
  signal cmd_local_reg_addr            : unsigned(REGABITS-1 downto 0);
  signal cmd_local_reg_addr_read       : unsigned(REGABITS-1 downto 0);
  signal cmd_reg_size                  : unsigned(4 downto 0);
  signal cmd_dev_addr                  : std_logic_vector(6 downto 0);
  signal cmd_reg_dev_addr              : std_logic_vector(6 downto 0);
  signal cmd_reg_addr                  : std_logic_vector(7 downto 0);
  signal cmd_dev_from_pointer          : std_logic;
  signal cmd_skip_cmd                  : std_logic;
  signal cmd_add_stop                  : std_logic;
  signal cmd_use_literal               : std_logic;
  signal cmd_address                   : unsigned(MEMABITS-1 downto 0);
  signal cmd_time                      : std_logic_vector(9 downto 0);
  signal cmd_ms                        : std_logic;
  signal cmd_src_a                     : unsigned(REGABITS-1 downto 0);
  signal cmd_dst_a                     : unsigned(REGABITS-1 downto 0);
  signal cmd_literal                   : std_logic_vector(7 downto 0);

  -- multi-byte transaction counter
  signal size_counter : unsigned(4 downto 0);

  -- adjust for big- and little-endian variants
  signal size_counter_endianness_adj : unsigned(4 downto 0);

  -- command arguments adjusted for endianness for multi-byte transactions
  signal cmd_local_reg_addr_or_literal_adj           : std_logic_vector(REGABITS-1 downto 0);
  signal cmd_local_reg_addr_adj                      : unsigned(REGABITS-1 downto 0);
  signal cmd_local_reg_addr_or_literal_adj_untrimmed : std_logic_vector(7 downto 0);
  signal cmd_local_reg_addr_adj_untrimmed            : unsigned(REGABITS-1 downto 0);

  -- holds data to write to the bus
  signal data_to_write : std_logic_vector(7 downto 0);

  -- wait timer control signals
  signal timer_start : std_logic;
  signal timer_count_ms : std_logic;
  signal timer_how_long : std_logic_vector(9 downto 0);
  signal timer_done : std_logic;

  -- register file control signals
  signal rf_start : std_logic;
  signal rf_write : std_logic;
  signal rf_addr  : std_logic_vector(REGABITS-1 downto 0);
  signal rf_din   : std_logic_vector(WORDWIDTH-1 downto 0);
  signal rf_dout  : std_logic_vector(WORDWIDTH-1 downto 0);
  signal rf_done  : std_logic;

  -- holds increment operation result
  signal inc_result : unsigned(WORDWIDTH-1 downto 0);

  -- helper counter for the syncio operation
  signal syncio_counter : unsigned(REGABITS-1 downto 0);

  -- error signals
  signal err_decode : std_logic;
  signal err_noack : std_logic;

begin

  -- instantiate the program memory
  gen_std_progmem: if SIMPLE_PROGMEM generate
    mouseq_progmem_1: entity work.mouseq_progmem
      port map (
        clk_i   => clk_i,
        rst_n_i => rst_n_i,
        start_i => pmem_fetch,
        addr_i  => pc_slv,
        data_o  => pmem_dout,
        done_o  => pmem_dv
        );
  end generate gen_std_progmem;


  gen_alt_progmem: if not SIMPLE_PROGMEM generate
    mouseq_sreg_progmem_inst: entity work.mouseq_sreg_progmem
      port map (
        clk_i   => clk_i,
        rst_n_i => rst_n_i,
        start_i => pmem_fetch,
        addr_i  => pc_slv,
        data_o  => pmem_dout,
        done_o  => pmem_dv
        );
  end generate gen_alt_progmem;


  pc_slv <= std_logic_vector(pc);


  -- populate the aliases used in all the various commands
  cmd_opcode                    <= cmd_word(15 downto 12);
  cmd_local_reg_addr_or_literal <= cmd_word(12 downto 5);
  cmd_local_reg_addr_read       <= unsigned(cmd_word(7+REGABITS-1 downto 7));
  cmd_local_reg_addr            <= unsigned(cmd_word(6+REGABITS-1 downto 6));
  cmd_reg_size                  <= unsigned(cmd_word(4 downto 0));
  cmd_dev_addr                  <= cmd_word_2(15 downto 9);
  cmd_reg_addr                  <= cmd_word_2(8 downto 1);
  cmd_add_stop                  <= cmd_word_2(0);
  cmd_dev_from_pointer          <= cmd_word(5);
  cmd_skip_cmd                  <= cmd_word(6);
  cmd_use_literal               <= '1' when (cmd_reg_size = "11111") else '0';
  cmd_address                   <= unsigned(cmd_word(4+MEMABITS-1 downto 4));
  cmd_time                      <= cmd_word(11 downto 2);
  cmd_ms                        <= cmd_word(1);
  cmd_src_a                     <= unsigned(cmd_word(6+REGABITS-1 downto 6));
  cmd_dst_a                     <= unsigned(cmd_word(REGABITS-1 downto 0));
  cmd_literal                   <= cmd_word_2(15 downto 8);


  -- helper signals for big- and little-endianness variants
  gen_big_endian_adj: if BIG_ENDIAN generate
    size_counter_endianness_adj <= size_counter;
  end generate gen_big_endian_adj;

  gen_little_endian_adj: if not BIG_ENDIAN generate
    size_counter_endianness_adj <= cmd_reg_size - size_counter;
  end generate gen_little_endian_adj;


  -- some gymnastics for width matching (thanks VHDL)
  process(clk_i) is
  begin
    if rising_edge(clk_i) then
      if (rst_n_i = '0') then
        cmd_local_reg_addr_or_literal_adj_untrimmed <= (others => '0');
        cmd_local_reg_addr_adj_untrimmed <= (others => '0');
      else
        cmd_local_reg_addr_or_literal_adj_untrimmed <= std_logic_vector(unsigned(cmd_local_reg_addr_or_literal) + size_counter_endianness_adj);
        cmd_local_reg_addr_adj_untrimmed <= cmd_local_reg_addr_read + size_counter_endianness_adj;
      end if;
    end if;
  end process;
  cmd_local_reg_addr_or_literal_adj <= cmd_local_reg_addr_or_literal_adj_untrimmed(REGABITS-1 downto 0);
  cmd_local_reg_addr_adj <= cmd_local_reg_addr_adj_untrimmed(REGABITS-1 downto 0);

  -- optionally add a register layer for the I/O
  io_regs : if BUFFERIO generate
    io_regs_proc : process (clk_i) is
    begin
      if rising_edge(clk_i) then        -- rising clock edge
        if (rst_n_i = '0') then         -- synchronous reset (active low)
          regs_in <= (others => '0');
          regs_out_o <= (others => '0');
        else
          regs_in <= regs_in_i;
          regs_out_o <= regs_out;
        end if;
      end if;
    end process io_regs_proc;
  end generate io_regs;


  -- or don't
  no_io_regs : if not BUFFERIO generate
    regs_in <= regs_in_i;
    regs_out_o <= regs_out;
  end generate no_io_regs;


  -- optionally add a watchdog
  watchdog: if WATCHDOG_ON generate
    -- purpose: reset the main fsm and the i2c master if the watchdog times out
    -- type   : sequential
    -- inputs : clk_i, rst_n_i, state
    -- outputs: rst
    watchdog_proc: process (clk_i) is
    begin  -- process watchdog_proc
      if rising_edge(clk_i) then        -- rising clock edge
        if rst_n_i = '0' then           -- synchronous reset (active low)
          rst     <= '1';
          state_r <= IDLE;

          watchdog_state <= WD_STATE_CHANGE;
          watchdog_cnt <= (others => '0');
        else
          rst     <= '0';
          state_r <= state;

          case watchdog_state is
            when WD_STATE_CHANGE =>
              watchdog_cnt <= (others => '0');
              if ((state_r = state) and
                  (state /= EXEC_WAIT_2) and
                  (state /= STARTUP)) then
                watchdog_state <= WD_STATE_SAME;
              end if;

            when WD_STATE_SAME =>
              watchdog_cnt <= watchdog_cnt + 1;
              if (watchdog_cnt = WATCHDOG_VAL) then
                watchdog_cnt <= (others => '0');
                watchdog_state <= WD_RESET;
              elsif ((state_r = state) and
                     (state /= EXEC_WAIT_2) and
                     (state /= STARTUP)) then
                watchdog_state <= WD_STATE_SAME;
              else
                watchdog_state <= WD_STATE_CHANGE;
              end if;

            when WD_RESET =>
              rst <= '1';
              watchdog_cnt <= watchdog_cnt + 1;
              if (watchdog_cnt = WATCHDOG_RESET_CYCLES) then
                watchdog_cnt <= (others => '0');
                watchdog_state <= WD_STATE_CHANGE;
              end if;

            when others =>
              watchdog_state <= WD_STATE_CHANGE;

          end case;

        end if;
      end if;
    end process watchdog_proc;
  end generate watchdog;

  -- purpose: generate the start_at_reset_addr flag for watchdog resets
  --          (rst_n_i will always make it start at pc=0)
  -- type : sequential inputs : clk_i, state outputs:
  -- start_at_reset_addr
  start_at_reset_addr_flag_proc: process (clk_i) is
  begin  -- process start_at_reset_addr_flag_proc
    if rising_edge(clk_i) then          -- rising clock edge
      if (rst_n_i = '0') then
        start_at_reset_addr1 <= '0';
        start_at_reset_addr2 <= '0';
        start_at_reset_addr3 <= '0';
      else
        start_at_reset_addr1 <= start_at_reset_addr_vote_out;
        start_at_reset_addr2 <= start_at_reset_addr_vote_out;
        start_at_reset_addr3 <= start_at_reset_addr_vote_out;
        if (state = FETCH_av) then
          start_at_reset_addr1 <= '1';
          start_at_reset_addr2 <= '1';
          start_at_reset_addr3 <= '1';
        end if;
      end if;
    end if;
  end process start_at_reset_addr_flag_proc;

  start_at_reset_addr_vote_out <= vote(start_at_reset_addr1,
                                       start_at_reset_addr2,
                                       start_at_reset_addr3);


  rst_inv: if not WATCHDOG_ON generate
    -- negate the reset to use in the i2c master and main FSM
    rst_inv_proc: process (clk_i) is
    begin  -- process watchdog_proc
      if rising_edge(clk_i) then        -- rising clock edge
        rst <= not rst_n_i;
      end if;
    end process rst_inv_proc;
  end generate rst_inv;


  -- instantiate the i2c master
  cmp_i2c_master : entity work.i2c_master_byte_ctrl
    port map (
      clk      => clk_i,
      rst      => rst,
      nReset   => '1',                  -- disable async reset
      ena      => i2c_en,
      clk_cnt  => I2C_CLKS,
      start    => i2c_start,
      stop     => i2c_stop,
      read     => i2c_rd,
      write    => i2c_wr,
      ack_in   => i2c_txack,

      i2c_busy => i2c_busy,
      i2c_al   => i2c_lost,
      din      => i2c_txbyte,
      cmd_ack  => i2c_done,
      ack_out  => i2c_rxack_n,
      dout     => i2c_rxbyte,

      scl_i    => scl_i,
      scl_o    => scl_o,
      scl_oen  => scl_oen,
      sda_i    => sda_i,
      sda_o    => sda_o,
      sda_oen  => sda_oen
      );


  -- instantiate the timer module
  mouseq_timer_inst: entity work.mouseq_timer
    generic map (
      CLK_PERIOD       => CLK_PERIOD,
      DISABLE_MS       => ESSENTIALS_ONLY,
      TOTAL_COUNT_BITS => 10)
    port map (
      clk_i    => clk_i,
      rst_n_i  => rst_n_i,
      start    => timer_start,
      count_ms => timer_count_ms,
      how_long => timer_how_long,
      done     => timer_done
      );


  -- instantiate the register file
  gen_sreg_regfile: if not SIMPLE_REGFILE generate
    mouseq_sreg_reg_file_inst: entity work.mouseq_sreg_reg_file
      port map (
        clk_i   => clk_i,
        rst_n_i => rst_n_i,
        start_i => rf_start,
        write_i => rf_write,
        addr_i  => rf_addr,
        data_i  => rf_din,
        data_o  => rf_dout,
        done_o  => rf_done
        );
  end generate gen_sreg_regfile;


  gen_simple_regfile: if SIMPLE_REGFILE generate
    mouseq_simple_reg_file_inst: entity work.mouseq_simple_reg_file
      port map (
        clk_i   => clk_i,
        rst_n_i => rst_n_i,
        start_i => rf_start,
        write_i => rf_write,
        addr_i  => rf_addr,
        data_i  => rf_din,
        data_o  => rf_dout,
        done_o  => rf_done
        );
  end generate gen_simple_regfile;


  -- purpose: implement main FSM
  -- type   : sequential
  -- inputs : cmd_add_stop, cmd_address, cmd_dev_addr,
  --          cmd_dst_a, cmd_literal, cmd_local_reg_addr,
  --          cmd_local_reg_addr_adj, cmd_local_reg_addr_or_literal,
  --          cmd_local_reg_addr_or_literal_adj, cmd_opcode,
  --          cmd_reg_addr, cmd_reg_size, cmd_src_a, cmd_time,
  --          cmd_ms, cmd_use_literal, i2c_done, i2c_rxack_n,
  --          i2c_rxbyte, pmem_dout, pmem_dv, regs_in, rf_done,
  --          rf_dout, rst, timer_done, start_at_reset_addr_vote_out
  -- outputs: state, data_to_write, size_counter, timer_start,
  --          timer_count_ms, timer_how_long, i2c_en, i2c_start,
  --          i2c_stop, i2c_rd, i2c_wr, i2c_txack, i2c_txbyte,
  --          rf_start, rf_write, rf_addr, rf_din, syncio_counter,
  --          err_decode, pc, cmd_word, cmd_word_2, i2c_rd,
  --          inc_result, regs_out, pmem_fetch, err_noack
  main_proc_fsm: process (clk_i) is
  begin  -- process main_proc_fsm
    if rising_edge(clk_i) then          -- rising clock edge
      if rst = '1' then             -- synchronous reset (active high)
        state <= IDLE;

        data_to_write <= (others => '0');

        size_counter  <= (others => '0');

        timer_start    <= '0';
        timer_count_ms <= '0';
        timer_how_long <= (others => '0');

        i2c_en     <= '0';
        i2c_start  <= '0';
        i2c_stop   <= '0';
        i2c_rd     <= '0';
        i2c_wr     <= '0';
        i2c_txack  <= '0';
        i2c_txbyte <= (others=>'0');

        rf_start <= '0';
        rf_write <= '0';
        rf_addr  <= (others => '0');
        rf_din   <= (others => '0');

        syncio_counter <= (others => '0');

        cmd_reg_dev_addr <= (others => '0');

        err_decode <= '0';
        err_noack  <= '0';

        pc         <= (others => '0');
        cmd_word   <= (others => '0');
        cmd_word_2 <= (others => '0');
        pmem_fetch <= '0';

        inc_result <= (others => '0');
        regs_out   <= (others => '0');
      else
        -- these signals should return to these states

        -- timer start signal should be pulsed, not held
        timer_start    <= '0';

        -- same with the rf_start
        rf_start <= '0';

        -- and with the decoding and noack errors
        err_decode <= '0';
        err_noack <= '0';

        -- and with that one
        pmem_fetch <= '0';

        -- i2c should be generally enabled
        i2c_en     <= '1';

        case (state) is
          when IDLE =>    -- IDLE state, we're here only after the reset
                          -- and maybe after errors?
            pc <= (others => '0');

            i2c_en     <= '0';
            i2c_start  <= '0';
            i2c_stop   <= '0';
            i2c_rd     <= '0';
            i2c_wr     <= '0';
            i2c_txack  <= '0';
            i2c_txbyte <= (others=>'0');

            timer_start <= '1';
            timer_how_long <= (0 => '1', others => '0');

            state <= STARTUP;
          when STARTUP =>          -- wait for the startup timer to ring
            if (start_at_reset_addr_vote_out = '1') then
              pc <= to_unsigned(RESET_START_FROM, MEMABITS);
            end if;

            if (timer_done = '1') then
              i2c_en <= '1';
              pmem_fetch <= '1';
              state <= FETCH_av;        -- don't increment the PC, first
                                        -- command
            end if;

          
          when FETCH =>
            pmem_fetch <= '1';
            pc <= pc + 1;
            state <= FETCH_av;

          when FETCH_av =>              -- memory address is correct
            if (pmem_dv = '1') then
              state <= FETCH_dv;
            end if;

          when FETCH_dv =>              -- memory results are correct
            cmd_word <= pmem_dout;
            state <= DECODE;

          when DECODE =>                -- decode the opcode
            case (cmd_opcode) is
              when c_op_WRITE1 =>
                pmem_fetch <= '1';
                pc <= pc + 1;
                state <= FETCH_WRITE_av; -- grab second word first
              when c_op_WRITE2 =>
                pmem_fetch <= '1';
                pc <= pc + 1;
                state <= FETCH_WRITE_av; -- grab second word first
              when c_op_READ1 =>
                pmem_fetch <= '1';
                pc <= pc + 1;
                state <= FETCH_READ_av; -- grab second word first
              when c_op_READ2 =>
                pmem_fetch <= '1';
                pc <= pc + 1;
                state <= FETCH_READ_av; -- grab second word first
              when c_op_STOP =>
                if ESSENTIALS_ONLY then -- command disabled
                  err_decode <= '1';
                  state <= FETCH;
                else
                  state <= EXEC_STOP;
                end if;
              when c_op_GOTO =>
                state <= EXEC_GOTO;
              when c_op_WAIT =>
                state <= EXEC_WAIT;
              when c_op_WAIT_TRIG =>
                state <= EXEC_WAIT_TRIG;
              when c_op_MOV =>
                if ESSENTIALS_ONLY then -- command disabled
                  err_decode <= '1';
                  state <= FETCH;
                else
                  state <= EXEC_MOV;
                end if;
              when c_op_SYNCIO =>
                state <= EXEC_SYNCIO;
              when c_op_LOAD =>
                if ESSENTIALS_ONLY then -- command disabled
                  err_decode <= '1';
                  state <= FETCH;
                else
                  pmem_fetch <= '1';
                  pc <= pc + 1;
                  state <= FETCH_LOAD_av;
                end if;
              when c_op_INC =>
                if ESSENTIALS_ONLY then -- command disabled
                  err_decode <= '1';
                  state <= FETCH;
                else
                  state <= EXEC_INC;
                end if;
              when c_op_NOP =>
                if ESSENTIALS_ONLY then -- command disabled
                  err_decode <= '1';
                  state <= FETCH;
                else
                  state <= FETCH;
                end if;
              when others =>
                err_decode <= '1';
                state <= FETCH;         -- shouldn't happen, throw error
            end case;

          when FETCH_WRITE_av =>        -- memory address is correct
            if (pmem_dv = '1') then
              state <= FETCH_WRITE_dv;
            end if;

          when FETCH_WRITE_dv =>        -- memory results are correct
            cmd_word_2 <= pmem_dout;    -- store second word, then exec
            state <= EXEC_WRITE;

          when FETCH_READ_av =>         -- memory address is correct
            if (pmem_dv = '1') then
              state <= FETCH_READ_dv;
            end if;

          when FETCH_READ_dv =>         -- memory results are correct
            cmd_word_2 <= pmem_dout;    -- store second word, then exec
            if (cmd_dev_from_pointer = '1') then
              state <= EXEC_READ_PNT;
            elsif (cmd_skip_cmd = '1') then
              state <= EXEC_READ_NOTFIRST;
            else
              state <= EXEC_READ;
            end if;
            

          when EXEC_WRITE =>            -- first issue a start condition
                                        -- and write the device address
            i2c_en     <= '1';
            i2c_start  <= '1';
            i2c_stop   <= '0';
            i2c_rd     <= '0';
            i2c_wr     <= '1';
            i2c_txack  <= '0';
            -- TODO: maybe we needed one more bit here?
            -- make sure a write address is selected
            i2c_txbyte <= cmd_dev_addr & '0';

            -- transaction over, go and write the register address
            if (i2c_done = '1') then
              i2c_start <= '0';
              i2c_wr    <= '0';
              err_noack <= i2c_rxack_n; -- no ack? error and continue
              state     <= EXEC_WRITE_REG;
            end if;

          when EXEC_WRITE_REG =>        -- write the register address
            size_counter <= cmd_reg_size;

            i2c_en     <= '1';
            i2c_start  <= '0';
            i2c_stop   <= '0';
            i2c_rd     <= '0';
            i2c_wr     <= '1';
            i2c_txack  <= '0';
            i2c_txbyte <= cmd_reg_addr;

            if (i2c_done = '1') then
              i2c_wr    <= '0';
              err_noack <= i2c_rxack_n; -- no ack? error and continue
              state     <= EXEC_WRITE_FETCH_NEXT_DATA; -- fetch the data
            end if;

          when EXEC_WRITE_DATA =>       -- write the data
            i2c_en     <= '1';
            i2c_start  <= '0';
            i2c_stop   <= '0';
            i2c_rd     <= '0';
            i2c_wr     <= '1';
            i2c_txack  <= '0';
            i2c_txbyte <= data_to_write;

            if (i2c_done = '1') then
              i2c_wr    <= '0';
              err_noack <= i2c_rxack_n; -- no ack? error and continue
              state     <= EXEC_WRITE_CHECK_SIZE;
            end if;

          when EXEC_WRITE_CHECK_SIZE => -- is this the last transaction?
            if ((cmd_use_literal = '1') or (size_counter = 0)) then  -- yes, it is
              if (cmd_add_stop='1') then
                state <= EXEC_WRITE_SEND_STOP;
              else
                state <= FETCH;
              end if;
            else                        -- no, it's not
              size_counter <= size_counter - 1;
              state <= EXEC_WRITE_WAIT_ADJ_UPD;
            end if;

          when EXEC_WRITE_WAIT_ADJ_UPD => -- endianness-adjusted address
                                          -- should be computed first
            state <= EXEC_WRITE_FETCH_NEXT_DATA;

          when EXEC_WRITE_FETCH_NEXT_DATA => -- select the data to use
            if (cmd_use_literal = '1') then  -- literal?
              data_to_write <= cmd_local_reg_addr_or_literal;

              state <= EXEC_WRITE_DATA;
            else                        -- or register?
              rf_start <= '1';
              rf_write <= '0';
              rf_addr  <= cmd_local_reg_addr_or_literal_adj;

              state <= EXEC_WRITE_WAIT_NEXT_DATA;
            end if;

          when EXEC_WRITE_WAIT_NEXT_DATA => -- wait for the reg file
            rf_start <= '0';
            rf_write <= '0';
            if (rf_done = '1') then
              data_to_write <= rf_dout;

              state <= EXEC_WRITE_DATA;
            end if;

          when EXEC_WRITE_SEND_STOP =>  -- send a stop condition
            i2c_en     <= '1';
            i2c_start  <= '0';
            i2c_stop   <= '1';
            i2c_rd     <= '0';
            i2c_wr     <= '0';
            i2c_txack  <= '0';
            i2c_txbyte <= data_to_write;

            if (i2c_done = '1') then
              i2c_wr    <= '0';
              i2c_stop  <= '0';
              err_noack <= i2c_rxack_n; -- no ack? error and continue
              state     <= FETCH;
            end if;

          when EXEC_READ =>             -- first issue a start condition
                                        -- and write the device address
            i2c_en     <= '1';
            i2c_start  <= '1';
            i2c_stop   <= '0';
            i2c_rd     <= '0';
            i2c_wr     <= '1';
            i2c_txack  <= '0';
            -- make sure a write address is selected
            i2c_txbyte <= cmd_dev_addr & '0';

            if (i2c_done = '1') then
              i2c_start <= '0';
              i2c_wr    <= '0';
              err_noack <= i2c_rxack_n; -- no ack? error and continue
              state     <= EXEC_READ_REG;
            end if;

          when EXEC_READ_PNT =>         -- register as pointer variant
            rf_start <= '1';
            rf_write <= '0';
            rf_addr  <= cmd_dev_addr(REGABITS-1 downto 0);
            state <= EXEC_READ_PNT_WAIT;

          when EXEC_READ_PNT_WAIT =>    -- wait for register output
            rf_start <= '0';
            rf_write <= '0';
            if (rf_done = '1') then
              cmd_reg_dev_addr <= rf_dout(6 downto 0);
              state <= EXEC_READ_PNT_GO;
            end if;

          when EXEC_READ_PNT_GO =>      -- like EXEC_READ
            i2c_en     <= '1';
            i2c_start  <= '1';
            i2c_stop   <= '0';
            i2c_rd     <= '0';
            i2c_wr     <= '1';
            i2c_txack  <= '0';
            -- make sure a write address is selected
            i2c_txbyte <= cmd_reg_dev_addr & '0';

            if (i2c_done = '1') then
              i2c_start <= '0';
              i2c_wr    <= '0';
              err_noack <= i2c_rxack_n; -- no ack? error and continue
              state     <= EXEC_READ_REG;
            end if;

          when EXEC_READ_REG =>         -- write the register address
            i2c_en     <= '1';
            i2c_start  <= '0';
            i2c_stop   <= '0';
            i2c_rd     <= '0';
            i2c_wr     <= '1';
            i2c_txack  <= '0';
            i2c_txbyte <= cmd_reg_addr;

            if (i2c_done = '1') then
              i2c_wr    <= '0';
              err_noack <= i2c_rxack_n; -- no ack? error and continue
              state     <= EXEC_READ_NOTFIRST;
            end if;

          when EXEC_READ_NOTFIRST =>    -- write the device address
                                        -- again
            size_counter <= cmd_reg_size;

            i2c_en     <= '1';
            i2c_start  <= '1';
            i2c_stop   <= '0';
            i2c_rd     <= '0';
            i2c_wr     <= '1';
            i2c_txack  <= '0';
            -- make sure a read address is selected this time
            if (cmd_dev_from_pointer = '1') then
              i2c_txbyte <= cmd_reg_dev_addr & '1';
            else
              i2c_txbyte <= cmd_dev_addr & '1';
            end if;

            if (i2c_done = '1') then
              i2c_start <= '0';
              i2c_wr    <= '0';
              err_noack <= i2c_rxack_n; -- no ack? error and continue
              state     <= EXEC_READ_DATA;
            end if;

          when EXEC_READ_DATA =>
            i2c_en     <= '1';
            i2c_start  <= '0';
            i2c_stop   <= '0';
            i2c_rd     <= '1';
            i2c_wr     <= '0';
            if (size_counter = 0) then  -- issue an ack unless it's the
                                        -- last word
              i2c_txack  <= '1';
            else
              i2c_txack  <= '0';
            end if;
            i2c_txbyte <= (others => '0');

            if (i2c_done = '1') then
              rf_start <= '1';          -- store data in the reg file
              rf_write <= '1';
              rf_addr  <= std_logic_vector(cmd_local_reg_addr_adj);
              rf_din   <= i2c_rxbyte;

              i2c_rd    <= '0';
              state <= EXEC_READ_STORE_DATA;
            end if;

          when EXEC_READ_STORE_DATA =>  -- wait for the reg file
            rf_start <= '0';
            rf_write <= '0';
            if (rf_done = '1') then
              state <= EXEC_READ_CHECK_SIZE;
            end if;

          when EXEC_READ_CHECK_SIZE =>  -- is this the last transaction?
            if (size_counter = 0) then  -- yes, it is
              if (cmd_add_stop='1') then
                i2c_txack  <= '0';
                state <= EXEC_READ_SEND_STOP;
              else
                state <= FETCH;
              end if;
            else                        -- no, it's not
              size_counter <= size_counter - 1;
              state <= EXEC_READ_WAIT_ADJ_UPD;
            end if;

          when EXEC_READ_WAIT_ADJ_UPD => -- endianness-adjusted address
                                         -- should be computed first
            state <= EXEC_READ_DATA;

          when EXEC_READ_SEND_STOP =>   -- send the stop
            i2c_en     <= '1';
            i2c_start  <= '0';
            i2c_stop   <= '1';
            i2c_rd     <= '0';
            i2c_wr     <= '0';
            i2c_txack  <= '0';
            i2c_txbyte <= (others => '0');

            if (i2c_done = '1') then
              i2c_stop   <= '0';
              state <= FETCH;
            end if;

          when EXEC_STOP =>
            i2c_stop   <= '1';

            state <= EXEC_STOP_2;

          when EXEC_STOP_2 =>           -- wait for the transaction to
                                        -- finish
            i2c_stop   <= '0';

            if (i2c_done = '1') then
              i2c_wr    <= '0';
              i2c_stop  <= '0';
              err_noack <= i2c_rxack_n; -- no ack? error and continue
              state     <= FETCH;
            end if;

          when EXEC_GOTO =>             -- the jump is simple
            pmem_fetch <= '1';
            pc <= cmd_address;
            state <= FETCH_av;

          when EXEC_WAIT =>          -- wait, in microseconds
            i2c_en      <= '0';

            timer_start <= '1';
            timer_count_ms <= cmd_ms;
            timer_how_long <= cmd_time;

            state <= EXEC_WAIT_2;

          when EXEC_WAIT_2 =>           -- wait for the timer
            i2c_en <= '0';

            if (timer_done = '1') then
              i2c_en <= '1';
              state <= FETCH;
            end if;

          when EXEC_WAIT_TRIG =>         -- wait for an external trigger
            if (ext_trig_i = '1') then
              state <= FETCH;
            end if;

          when EXEC_MOV =>              -- fetch the data first
            rf_start <= '1';
            rf_write <= '0';
            rf_addr <= std_logic_vector(cmd_src_a);
            state <= EXEC_MOV_WAIT_R;

          when EXEC_MOV_WAIT_R =>       -- wait for the reg file
            rf_start <= '0';
            rf_write <= '0';
            if (rf_done = '1') then     -- write the data now
              rf_start <= '1';
              rf_write <= '1';
              rf_din   <= rf_dout;
              rf_addr  <= std_logic_vector(cmd_dst_a);
              state <= EXEC_MOV_WAIT_W;
            end if;

          when EXEC_MOV_WAIT_W =>       -- wait for the reg file
            rf_start <= '0';
            rf_write <= '0';
            if (rf_done = '1') then
              state <= FETCH;
            end if;

          when FETCH_LOAD_av =>         -- memory address is correct
            if (pmem_dv = '1') then
              state <= FETCH_LOAD_dv;
            end if;

          when FETCH_LOAD_dv =>         -- memory results are correct
            cmd_word_2 <= pmem_dout;    -- store second word, then exec
            state <= EXEC_LOAD;

          when EXEC_LOAD =>             -- load to a register
            rf_start <= '1';
            rf_write <= '1';
            rf_addr <= std_logic_vector(cmd_local_reg_addr);
            rf_din <= cmd_literal;
            state <= EXEC_LOAD_WAIT;

          when EXEC_LOAD_WAIT =>        -- wait for the reg file
            rf_start <= '0';
            rf_write <= '0';
            if (rf_done = '1') then
              state <= FETCH;
            end if;

          when EXEC_INC =>              -- request the initial value
            rf_start <= '1';
            rf_write <= '0';
            rf_addr <= std_logic_vector(cmd_local_reg_addr);
            state <= EXEC_INC_WAIT_R;

          when EXEC_INC_WAIT_R =>       -- wait for the reg file
            rf_start <= '0';
            rf_write <= '0';
            if (rf_done = '1') then     -- got it, increment
              inc_result <= unsigned(rf_dout) + 1;
              state <= EXEC_INC_INC;
            end if;

          when EXEC_INC_INC =>          -- write the result
            rf_start <= '1';
            rf_write <= '1';
            rf_din   <= std_logic_vector(inc_result);

            state <= EXEC_INC_WAIT_W;

          when EXEC_INC_WAIT_W =>       -- wait for the reg file
            rf_start <= '0';
            rf_write <= '0';
            if (rf_done = '1') then
              state <= FETCH;
            end if;

          when EXEC_SYNCIO =>           -- first synchronize input regs
            if (syncio_counter < IWORDS) then
              rf_start <= '1';
              rf_write <= '1';
              rf_addr  <= std_logic_vector(syncio_counter);
              rf_din   <= regs_in((to_integer(syncio_counter)+1)*WORDWIDTH-1 downto
                                  to_integer(syncio_counter)*WORDWIDTH);

              state <= EXEC_SYNCIO_READIN_WAIT;
            else
              state <= EXEC_SYNCIO_WRITEOUT;
            end if;

          when EXEC_SYNCIO_READIN_WAIT => -- wait for reg file
            rf_start <= '0';
            rf_write <= '0';

            if (rf_done = '1') then     -- next register
              syncio_counter <= syncio_counter + 1;

              state <= EXEC_SYNCIO;
            end if;

          when EXEC_SYNCIO_WRITEOUT =>  -- then do the output regs
            if (syncio_counter < (IWORDS + OWORDS)) then
              rf_start <= '1';
              rf_write <= '0';
              rf_addr  <= std_logic_vector(syncio_counter);

              state <= EXEC_SYNCIO_WRITEOUT_WAIT;
            else
              syncio_counter <= (others => '0');
              state <= FETCH;
            end if;

          when EXEC_SYNCIO_WRITEOUT_WAIT => -- wait for the reg file
            rf_start <= '0';
            rf_write <= '0';

            if (rf_done = '1') then
              regs_out((to_integer(syncio_counter)-IWORDS+1)*WORDWIDTH-1 downto
                (to_integer(syncio_counter)-IWORDS)*WORDWIDTH) <= rf_dout;

              syncio_counter <= syncio_counter + 1;

              state <= EXEC_SYNCIO_WRITEOUT;
            end if;

          when others =>                -- TODO: put something here for
                                        -- safe implementations?
            state <= IDLE;

        end case;
        
      end if;
    end if;
  end process main_proc_fsm;

end behavioral;
