library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use ieee.math_real.all;
use std.textio.all;

package mouseq_pkg is

  constant IWORDS : natural := 0; -- input registers
  constant OWORDS : natural := 47; -- output registers
  constant INTWORDS : natural := 0;    -- internal registers

  constant I2C_CLKS : unsigned(15 downto 0) := x"0080"; -- i2c clock prescale

  constant MEMWIDTH : natural := 16;               -- program mem word width, fixed
  constant WORDWIDTH : natural := 8;               -- reg width, fixed

  -- program memory size in words, address bus size
  constant MEMSIZE : natural := 64;
  constant MEMABITS : natural := integer(ceil(log2(real(MEMSIZE))));

  -- define a type for the program memory
  subtype RamCellType is std_logic_vector(MEMWIDTH-1 downto 0);
  type RamType is array(0 to MEMSIZE-1) of RamCellType;

  -- program memory initialization file
  constant PROG_FILENAME : string := "mouseq_mem.mif";

  -- register file size in words, address bus size
  constant REGSIZE : natural := IWORDS+OWORDS+INTWORDS;
  constant REGABITS : natural := integer(ceil(log2(real(REGSIZE))));

  -- define a type for the unified register space
  type RegType is array(0 to IWORDS+OWORDS+INTWORDS-1) of std_logic_vector(WORDWIDTH-1 downto 0);

  constant BUFFERIO : boolean := false; -- add register layer I/O or not
  constant BIG_ENDIAN : boolean := false; -- select endianness for multi-byte
  constant FAMILY : string := "NX"; -- select device family for mem implementation
  constant SIMPLE_PROGMEM : boolean := false; -- select simple prog memory implementation (more resources with TMR)
  constant SIMPLE_REGFILE : boolean := false; -- select simple register file implementation (more resources)
  constant TMR : boolean := false; -- TMR constant used to set attributes
  constant ESSENTIALS_ONLY : boolean := false; -- disable NOP, MOV, INC, LOAD, STOP and WAIT_MS

  constant WATCHDOG_ON : boolean := false; -- enable watchdog
  constant WATCHDOG_VAL : natural := 1; -- watchdog value
  constant WATCHDOG_BITS : natural := integer(ceil(log2(real(WATCHDOG_VAL))));
  constant WATCHDOG_RESET_CYCLES : natural := 10; -- how many cycles to keep the watchdog reset high

  constant RESET_START_FROM : natural := 0; -- reset at a certain address to skip init

  -- define a helper function to initialize the program memory
  impure function InitRamFromFile(RamFileName : in string) return RamType;
  impure function InitRamFirstCellFromFile(RamFileName : in string) return RamCellType;

  function vote(I1 : std_logic; I2 : std_logic; I3 : std_logic) return std_logic;
  function get_attr_radhard return string;

end package mouseq_pkg;

package body mouseq_pkg is

  impure function InitRamFromFile(RamFileName : in string) return RamType is
    file RamFile : text is in RamFileName;
    variable RamFileLine : line;
    variable temp_bv : bit_vector(MEMWIDTH-1 downto 0);
    variable RAM : RamType;
  begin
    for I in RamType'range loop
      readline(RamFile, RamFileLine);
      read(RamFileLine, temp_bv);
      RAM(I) := to_stdlogicvector(temp_bv);
    end loop;
    return RAM;
  end function;

  impure function InitRamFirstCellFromFile(RamFileName : in string) return RamCellType is
    file RamFile : text is in RamFileName;
    variable RamFileLine : line;
    variable temp_bv : bit_vector(MEMWIDTH-1 downto 0);
  begin
    readline(RamFile, RamFileLine);
    read(RamFileLine, temp_bv);
    return(to_stdlogicvector(temp_bv));
  end function;

  -- purpose: manually implement TMR to scrub memories
  function vote(I1 : std_logic; I2 : std_logic; I3 : std_logic)
    return std_logic is
  begin  -- function vote
    return ((I1 and I2) or (I2 and I3) or (I3 and I1));
  end function vote;

  -- purpose: set the syn_radhardlevel attribute if TMR is set
  function get_attr_radhard return string is
  begin
    if TMR then
      return "tmr";
    else
      return "none";
    end if;
  end function get_attr_radhard;

end package body mouseq_pkg;
