/*
 * Copyright (C) 2018 CERN
 * Author: Maciej Suminski <maciej.suminski@cern.ch>
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "diot_crate.h"
#include <iostream>
#include <iomanip>
#include <limits>
#include <cmath>

void DIOT_Crate::SetOutput(int slot, int idx, bool state)
{
    const int var_offset = getDataOffset(slot, idx);
    const uint8_t io_mask = getIOMask(idx);
    uint8_t data = GetProdVarData(var_offset);

    if ((data & io_mask) == state)
        return;     // no update needed

    if (state)
        data |= io_mask;
    else
        data &= ~io_mask;

    SetProdVarData(var_offset, data);
}


void DIOT_Crate::SetSlotOutputs(int slot, uint32_t state)
{
    const int var_offset = getDataOffset(slot, 0);

    {   // mutex locked scope
    auto lock = getProdLock();
    setProdDataLock(var_offset, (uint8_t)(state & 0xff), lock);
    setProdDataLock(var_offset+1, (uint8_t)(state >> 8 & 0xff), lock);
    setProdDataLock(var_offset+2, (uint8_t)(state >> 16 & 0xff), lock);
    setProdDataLock(var_offset+3, (uint8_t)(state >> 24 & 0xff), lock);
    }
}


bool DIOT_Crate::GetFanStatus(unsigned int fan)
{
    if (fan >= FAN_COUNT)
        return false;

    int status = GetFanStatusAll();
    // status bit == 1 indicates failure, hence negation
    return !(status & (1 << fan));
}


int DIOT_Crate::GetFanRPM(unsigned int fan)
{
    if (fan >= FAN_COUNT)
        return std::numeric_limits<int>::min();

    int rpm = 0;

    {   // mutex locked scope
    auto lock = getConsLock();
    rpm = (int) getConsDataLock(FAN0_SPEED + fan * 2, lock) << 8;
    rpm |= getConsDataLock(FAN0_SPEED + fan * 2 + 1, lock);
    }

    return rpm;
}


float conv_direct_to_float(unsigned int lin_int)
{
    int mantissa = 0;
    int exponent = 0;

    if (lin_int & 0x400)
      mantissa = (int)(0xfffff800 | (lin_int & 0x7ff));
    else
      mantissa = (int)(lin_int & 0x3ff);

    if (lin_int & 0x00008000)
      exponent = (int)(0xffffffe0 | (lin_int & 0xf800) >> 11);
    else
      exponent = (int)((lin_int & 0xf800) >> 11);

    return(mantissa*std::pow(2.0, exponent));
}


float DIOT_Crate::GetPSUVoltage(unsigned int psu)
{
    if (psu >= 2)
        return std::numeric_limits<int>::min();

    int volts = 0;
    int psuvoutmode = 0;

    {   // mutex locked scope
    auto lock = getConsLock();
    volts = (int) getConsDataLock(PSUVOLT1 + psu * 2 + 1, lock) << 8;
    volts |= getConsDataLock(PSUVOLT1 + psu * 2, lock);
    psuvoutmode = getConsDataLock(PSUVOUTMODE1 + psu, lock);
    if (psuvoutmode & 0x10)
      psuvoutmode = (0x1f - psuvoutmode) + 1;
    }

    return volts / std::pow(2, psuvoutmode);
}


float DIOT_Crate::GetPSUCurrent(unsigned int psu)
{
    if (psu >= 2)
        return std::numeric_limits<int>::min();

    int curr = 0;

    {   // mutex locked scope
    auto lock = getConsLock();
    curr = (int) getConsDataLock(PSUCURR1 + psu * 2 + 1, lock) << 8;
    curr |= getConsDataLock(PSUCURR1 + psu * 2, lock);
    }

    return conv_direct_to_float(curr);
}


float DIOT_Crate::GetPSUPower(unsigned int psu)
{
    if (psu >= 2)
        return std::numeric_limits<int>::min();

    int power = 0;
    int mode = 0;
    int vout_p1 = 0;
    int vout_p2 = 0;

    {   // mutex locked scope
    auto lock = getConsLock();
    power = (int) getConsDataLock(PSUPOWER1 + psu * 2 + 1, lock) << 8;
    power |= getConsDataLock(PSUPOWER1 + psu * 2, lock);
    mode   = getConsDataLock(PSUVOUTMODE1 + psu, lock);
    vout_p1 = (int) getConsDataLock(PSUVOLT_P1_1 + psu * 4 + 1, lock) << 8;
    vout_p1 |= getConsDataLock(PSUVOLT_P1_1 + psu * 4, lock);
    vout_p2 = (int) getConsDataLock(PSUVOLT_P2_1 + psu * 4 + 1, lock) << 8;
    vout_p2 |= getConsDataLock(PSUVOLT_P2_1 + psu * 4, lock);
    }

    std::cout << vout_p1;
    std::cout << vout_p2;

    std::cout << " mode: " << mode << std::endl;
    std::cout << " vout p1: " << vout_p1 / std::pow(2, 9) << std::endl;
    std::cout << " vout p2: " << vout_p2 / std::pow(2, 9) << std::endl;
    return conv_direct_to_float(power);
}


float DIOT_Crate::GetPSUTemperature(unsigned int psu)
{
    if (psu >= 2)
        return std::numeric_limits<int>::min();

    int temp = 0;

    {   // mutex locked scope
    auto lock = getConsLock();
    temp = (int) getConsDataLock(PSUTEMP1 + psu * 2 + 1, lock) << 8;
    temp |= getConsDataLock(PSUTEMP1 + psu * 2, lock);
    }

    return conv_direct_to_float(temp);
}


bool DIOT_Crate::GetTempStatus(unsigned int sensor)
{
    if (sensor >= TEMP_COUNT)
        return false;

    int status = GetTempStatusAll();
    // status bit == 1 indicates failure, hence negation
    return !(status & (1 << sensor));
}


int DIOT_Crate::GetTemp(unsigned int sensor)
{
    if (sensor >= TEMP_COUNT)
        return std::numeric_limits<int>::min();

    return GetConsVarData(TEMPERATURE0 + sensor);
}


std::string DIOT_Crate::GetVoltName(VOLTAGE volt)
{
    for (const auto& volt_pair : VOLTAGES) {
        if (volt_pair.first == volt)
            return volt_pair.second;
    }

    return "UNKNOWN";
}


DIOT_Crate::VOLTAGE DIOT_Crate::GetVoltEnum(unsigned int idx)
{
    if (idx >= VOLTAGES.size())
        return INVALID;

    return VOLTAGES[idx].first;
}


uint32_t DIOT_Crate::getConsWord(int address)
{
    uint32_t ret = 0;

    {   // mutex locked scope
    auto lock = getConsLock();

    for (int i = 3; i >= 0; --i) {
        ret <<= 8;
        ret |= getConsDataLock(address + i, lock);
    }
    }

    return ret;
}


const std::vector<std::pair<DIOT_Crate::VOLTAGE, std::string>> DIOT_Crate::VOLTAGES =
    { { P3V3, "+3.3V" }, { P5V0, "+5V" },
      { P12V0, "+12V" }, { M12V0, "-12V" } };
